/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2007 Smithsonian Astrophysical Observatory
 *
 * This file is part of rdbplt
 *
 * rdbplt is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * rdbplt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdlib.h>
#include <stdio.h>

#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>

#include <suplib/str.h>
#include <suplib/debug.h>
#include <tparam/tparam.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

static TokListToken
axis_types[] = 
{
  { "linlin",	Axis_LINLIN },
  { "linlog",	Axis_LINLOG },
  { "loglin",	Axis_LOGLIN },
  { "loglog",	Axis_LOGLOG },
};

static TokList axis_type = GenTokList(axis_types);

/* limit algorithm */
static TokListToken
limit_alg[] =
{
  {"none",	LimitAlg_NONE},
  {"percent",	LimitAlg_PERCENT},
  {"pgplot",	LimitAlg_PGPLOT},
};
static TokList limit_algs = GenTokList(limit_alg);


static double	parse_float( const char *opt, const char *value, int val );
static void	parse_legend_text( Legend *legend, char *text );
static void	parse_legend_opts( Legend *legend, char *opt_str );

static void	parse_error_bar(paramfile pf, char *param,
				char column[][STR_LENGTH+1],
				ErrorBarAttr eba_def[],
				ErrorBarAttr *eba[], int* n_eba);
static void	parse_color( paramfile pf );
static void	parse_markers( paramfile pf );
static void	parse_connect( paramfile pf );


void
gparam(int argc, char *argv[])
{
  paramfile pf;
  char dummy[1024];
  
  tf_enter("gparam");

  pf = tparamopen(NULL, argv, argc, "r");

  /* get debug flags */
  tpgetstr(pf, "debug", debug, DEBUG_LENGTH, 0);
  debug_init(debug);

  c.version = tpgetb( pf, "version" );
  c.help = tpgetb( pf, "help" );

  if ( c.version || c.help )
    goto leave;


  /* input files */
  tpgetstr(pf, "input", c.input, STR_LENGTH, 1);

  /* device */
  tpgetstr(pf, "device", c.device, STR_LENGTH, 1);

  /* annotations */
  /* axis labels will be set to the names of the columns if not
     specified. this is done below */
  tpgetstr(pf, "xlabel", c.xlabel, STR_LENGTH, 0);
  tpgetstr(pf, "ylabel", c.ylabel, STR_LENGTH, 0);
  tpgetstr(pf, "title",  c.title,  STR_LENGTH, 0);
  tpgetstr(pf, "subtitle",  c.subtitle,  STR_LENGTH, 0);
    
  /* break column */
  tpgetstr(pf, "break", c.break_col, STR_LENGTH, 1);
  if ( is_false( c.break_col ) )
    c.break_col[0] = '\0';
  else
    c.do_break = 1;

  {
    AxisXFRM axisxfrm;

    /* axis transformation */
    tpgetstr(pf, "axisxfrm", dummy, STR_LENGTH, 1);
    if ( -1 == ( axisxfrm = tokmatch( dummy, &axis_type ) ) )
      tf_exit( ExitERR_illegalArg, "unknown axisxfrm type: `%s'", dummy );

    c.do_log_x = c.do_log_y = 0;
    switch( axisxfrm )
    {
    case Axis_LINLIN:
      break;

    case Axis_LINLOG:
      c.do_log_y++;
      break;

    case Axis_LOGLIN:
      c.do_log_x++;
      break;

    case Axis_LOGLOG:
      c.do_log_x++;
      c.do_log_y++;
      break;

    default:
      tf_exit( ExitERR_misc, "internal error" );
    }
  }

  c.linewidth = tpgeti( pf, "line_width" );
  c.char_height = tpgetd( pf, "char_height" );

  /* now, set the defaults, using the new values of linewidth
     and char_height */

  set_def_attrs();

  /* flags */
  tpgetstr(pf, "autoscale", dummy, STR_LENGTH, 1);
  if ( 4 != strlen( dummy ) || 4 != strspn( dummy, "01" ) )
    tf_exit( ExitERR_illegalArg, "illegal autoscale value" );

  c.auto_xmin = '1' == dummy[0];
  c.auto_xmax = '1' == dummy[1];
  c.auto_ymin = '1' == dummy[2];
  c.auto_ymax = '1' == dummy[3];

  tpgetstr(pf, "axisflip", dummy, STR_LENGTH, 1);
  if ( 2 != strlen( dummy ) || 2 != strspn( dummy, "01" ) )
    tf_exit( ExitERR_illegalArg, "illegal axisflip value" );

  c.flip_x = '1' == dummy[0];
  c.flip_y = '1' == dummy[1];

  parse_color( pf );
  parse_connect( pf );
  parse_markers( pf );

  /* columns to plot */
  tpgetstr(pf, "xcol", c.xcol, STR_LENGTH, 1);
  tpgetstr(pf, "ycol", c.ycol, STR_LENGTH, 1);

  /* create default axis labels for the lazy user */
  if ( '\0' == c.xlabel[0] )
    strcpy( c.xlabel, c.xcol );
  if ( '\0' == c.ylabel[0] )
    strcpy( c.ylabel, c.ycol );

  /* error columns */
  parse_error_bar( pf, "xerr", c.xerr, c.xerr_attr_def,
		   c.xerr_attr, &c.n_xerr_attr );

  parse_error_bar( pf, "yerr", c.yerr, c.yerr_attr_def,
		   c.yerr_attr, &c.n_yerr_attr );

  c.xmax = tpgetd( pf, "xmax" );
  c.xmin = tpgetd( pf, "xmin" );
  c.ymax = tpgetd( pf, "ymax" );
  c.ymin = tpgetd( pf, "ymin" );

  tpgetstr(pf, "limit_alg", dummy, STR_LENGTH, 1);
  c.limit_alg = tokmatch(dummy, &limit_algs);
  if (c.limit_alg == -1)
    tf_exit(ExitERR_illegalArg, "unknown limit algorithm: `%s'", dummy);

  if (c.limit_alg == LimitAlg_PERCENT)
    c.limit_val = tpgetd(pf, "limit_val");

  c.just = tpgetb( pf, "just" );
  c.width = tpgetd( pf, "width" );
  c.height = tpgetd( pf, "height" );

  c.legend.plot = pgetb( pf, "legend" );

  if ( c.legend.plot )
  {
    tpgetstr(pf, "legend_opts", dummy, 1023, 0);

    /* split up legend_opts */
    parse_legend_opts( &c.legend, dummy );

    tpgetstr(pf, "legend_text", dummy, 1023, 0);
    parse_legend_text( &c.legend, dummy );

    if ( 0 == c.legend.n )
    {
      parse_legend_text( &c.legend, "%s" );
    }
  }

  {
    int n;

    int nattr = c.n_markers > c.n_connect ? c.n_markers : c.n_connect;

    int doit =
	   c.marker_attr_def.doit  != Marker_None
	|| c.connect_attr_def.doit != Connect_None
	;

    int do_ebars =
	   c.xerr_attr_def[0].doit
	|| c.xerr_attr_def[1].doit
	|| c.yerr_attr_def[0].doit
	|| c.yerr_attr_def[1].doit;

    doit |= do_ebars;

    if ( c.n_markers == 0 && c.n_connect == 0 && ! (doit || do_ebars) )
      tf_exit( ExitERR_illegalArg,
	       "must specify at least one of marker or connect or xerr or yerr" );

    for ( n = 0 ; n < nattr ; n++ )
    {
      int do_m = (c.n_markers > n ?
		  c.marker_attr[n].doit : c.marker_attr_def.doit)
	!= Marker_None;

      int do_c = (c.n_connect > n ?
		  c.connect_attr[n].doit : c.connect_attr_def.doit)
	!= Connect_None;

      if ( ! (do_m || do_c || do_ebars ) )
	tf_exit( ExitERR_illegalArg,
		 "must specify at least one of marker or connect "
		 "for dataset %d", n+1 );
    }
  }


 leave:
  paramclose(pf);
  tf_leave();
}

static void
parse_legend_opts( Legend *legend, char *opt_str )
{
  char *s;

  s = strtok( opt_str, "; " );
  while ( s )
  {
    int val = 0;
    char *v;

    /* look for a '=' */
    if ( v = strrchr( s, '=' ) )
    {
      val++;
      *v++ = '\0';
    }

    if      ( ! strcmp( s, "x" ) )
    {
      legend->x = parse_float( s, v, val );
    }
    else if ( ! strcmp( s, "y" ) )
    {
      legend->y = parse_float( s, v, val );
    }
    else if ( ! strcmp( s, "lineskip" ) )
    {
      legend->lineskip = parse_float( s, v, val );
    }
    else if ( ! strcmp( s, "linelen" ) )
    {
      legend->line_length = parse_float( s, v, val );
    }
    else if ( ! strcmp( s, "boxpad" ) )
    {
      legend->box_pad = parse_float( s, v, val );
    }
    else if ( ! strcmp( s, "ul" ) )
    {
      legend->pos = Legend_UL;
    }
    else if ( ! strcmp( s, "ll" ) )
    {
      legend->pos = Legend_LL;
    }
    else if ( ! strcmp( s, "box" ) )
    {
      if ( val )
	tf_exit( ExitERR_illegalArg,
		 "legend option `%s' doesn't take a value", s );

      legend->box++;
    }
    else if ( ! strcmp( s, "char_ht" ) )
    {
      legend->char_ht = parse_float( s, v, val );
    }
    else
    {
      tf_exit( ExitERR_illegalArg, "unknown legend option `%s'", s );
    }

    s = strtok( NULL, "; " );
  }
}

static double
parse_float( const char *opt, const char *value, int val )
{
  double d;
  char *p;

  if ( ! val )
    tf_exit( ExitERR_illegalArg, "legend option `%s' requires a value", opt );

  d = strtod( value, &p );
  if ( *p != '\0' || errno == ERANGE )
    tf_exit( ExitERR_illegalArg,
	     "legend option `%s' value `%s': illegal float", opt, value );

  return d;
}

static void
parse_legend_text( Legend *legend, char *text )
{
  char *s;
  int i;

  legend->n = str_tokcnt( text, "\n" );

  if ( 0 == legend->n )
    return;

  if ( NULL == ( legend->text = str_dup( text ) ) )
    tf_exit( ExitERR_alloc, "couldn't dup legend text" );

  if ( NULL ==
       ( legend->textp = malloc ( legend->n * sizeof( char * ) ) ) )
    tf_exit( ExitERR_alloc, "couldn't allocate legend text pointers" );

  s = strtok( legend->text, "\n" );
  for ( i = 0 ; i < legend->n ; i++ )
  {
    if ( ! s )
      tf_exit( ExitERR_misc, "%s %d: internal error", __FILE__, __LINE__ );

    legend->textp[i] = s;

    s = strtok( NULL, "\n" );
  }
}

static void
parse_error_bar(
  paramfile pf,
  char *param,
  char column[][STR_LENGTH+1],
  ErrorBarAttr eba_def[],
  ErrorBarAttr *eba[],
  int* n_eba
  )
{
  char dummy[STR_LENGTH+1];
  char* tok[2];
  int ntoks;
  char *col;

  char* attrp;

  tf_enter( "parse_error_bar" );

  tpgetstr(pf, param, dummy, STR_LENGTH, 0);

  strip_space( dummy );

  /* error bar specifications have one of the following forms

     err
     err:attr1_n%attr1_p,attr2,,attr3
     errp,errn
     errp,errn:attr1,attr2,,,attr3

     attributes are line attributes with an extra attribute of
     character height

  */

  /* if there's a colon, seal off the preceding characters
     deal with the default specs
  */

  if ( attrp = strchr( dummy, ':' ) )
    *attrp++ = '\0';

  /* grab column names */
  if ( (ntoks = toksplit( dummy, tok, ",", 2 )) > 2 )
    tf_exit( ExitERR_misc, "too many columns for `%s'", param );

  /* this is for the defaults */
  parse_ErrorBarAttr( tok[0], &col, &eba_def[0] );
  if ( col )
  {
    strcpy(column[0],col);
    free( col );
  }

  if ( ntoks == 1 )
    parse_ErrorBarAttr( tok[0], &col, &eba_def[1] );
  else
    parse_ErrorBarAttr( tok[1], &col, &eba_def[1] );

  if ( col )
  {
    strcpy(column[1],col);
    free( col );
  }

  /* if there were attributes for the individual data sets,
     deal with them */
  if ( attrp )
  {
    int i;
    char *ptr;
    char dchar;
    char *token;

    *n_eba = str_tokqcnt( attrp, ",", 0 );
    
    eba[0] = (ErrorBarAttr *) malloc( *n_eba * sizeof(ErrorBarAttr) );
    eba[1] = (ErrorBarAttr *) malloc( *n_eba * sizeof(ErrorBarAttr) );

    if ( NULL == eba[0] || NULL == eba[1] )
      tf_exit( ExitERR_alloc, "unable to allocate %d ErrorBarAttrs",
	       *n_eba );

    token = str_tokq( attrp, ",", 0, 1, &ptr, &dchar );

    for ( i = 0 ; token && i < *n_eba ;
	  i++, token = str_tokq( NULL, ",", 0, 1, &ptr, &dchar )
	  )
    {
      int ss;

      /* split on negative & positive attrs */
      ntoks = toksplit( token, tok, "%", 2 );
      if ( ntoks > 2 )
	tf_exit( ExitERR_misc,
		 "too many columns in dataset section for `%s'", param );

      /* if no attributes for this data set, use default. set values
	 to the empty string to defeat is_false, but so
	 parse_ErrorBarAttr can handle it correctly */
      if ( 0 == ntoks )
	tok[1] = tok[0] = "";

      /* repeat if only one specified */
      else if ( 1 == ntoks )
	tok[1] = tok[0];

      for ( ss = 0 ; ss < 2 ; ss++ )
      {
	eba[ss][i] = eba_def[ss];

	if ( is_false( tok[ss] ) )
	  eba[ss][i].doit = 0;

	else
	{
	  eba[ss][i].doit = 1;
	  parse_ErrorBarAttr( token, NULL, &eba[ss][i] );
	}
      }
    }

    tf_assert( i == *n_eba && token == NULL );

  }

  tf_leave();
}


static void
parse_color( paramfile pf )
{
  int i;
  char *token;
  char *buf;
  char dummy[STR_LENGTH+1];

  tf_enter( "parse_color" );
      
  buf = dummy;

  tpgetstr( pf, "color", buf, STR_LENGTH, 0 );

  /* strip trailing space */
  strip_space( buf );

  c.n_colors = 0;
  if ( is_false( buf ) )
    goto leave;

  /* color cycling? */
  c.color_cycle = 0;
  if ( 0 == strncmp( buf, "cycle", 5 ) )
  {
      c.color_cycle = 1;
      buf += 5;
      if ( buf[0] && ( buf[0] == ':' || buf[0] == ',' ) )
	  buf++;
  }

  c.n_colors = str_tokcnt( buf, "," );

  if ( 0 == c.n_colors )
    c.color = NULL;

  else
  {
    c.color = (int *) malloc( c.n_colors * sizeof(*c.color) );
    if ( NULL == c.color )
      tf_exit( ExitERR_alloc, "unable to allocate %d colors",
	       c.n_colors);

    for ( i = 0, token = strtok( buf, "," ) ;
	  token && i < c.n_colors;
	  token = strtok( NULL, "," ), i++ )
    {
      c.color[i] = tokmatch(token, &color_spec);
      if ( -1 == c.color[i] )
	tf_exit( ExitERR_illegalArg, "illegal color: %s", token );
    }
    tf_assert( i == c.n_colors && token == NULL );
  }

  if ( dbflag( "dump" ))
    dump_Colors( stderr );

 leave:
  tf_leave();
}

/*
  markers can have the form

  attr,attr,attr
  [%]colname[/attr][:attr,attr,...]



 */

static void
parse_markers( paramfile pf )
{
  char dummy[STR_LENGTH+1];
  char *dsetp;
  char *spec = dummy;

  tf_enter( "parse_markers" );
  
  tpgetstr( pf, "marker", dummy, STR_LENGTH, 1 );

  /* strip trailing space */
  strip_space( dummy );

  c.marker_col[0] = '\0';
  c.marker_attr = NULL;
  c.n_markers = 0;

  if ( is_true( spec ) )
  {
    c.marker_attr_def.doit = Marker_Symbol;
    goto leave;
  }
    
  if ( is_false( spec ) )
  {
    c.marker_attr_def.doit = Marker_None;
    goto leave;
  }

  /* if there's a colon, seal off the preceding characters
     deal with the default specs
  */
  if ( (dsetp = strchr( spec, ':' )) || (NULL == strchr( spec, ',')) )
  {
    char* res;
    char* col = NULL;
    MarkerDoWhat doit = Marker_Symbol;

    if (dsetp)
      *dsetp = '\0';

    if ( '%' == spec[0] )
    {
      c.marker_attr_def.doit = doit = Marker_ColString;
      spec++;
    }

    res = parse_MarkAttr( spec, &col, &c.marker_attr_def );

    if ( *res )
      tf_message(
	"ignoring extra attributes in global marker attribute spec `%s'\n",
	dummy );

    if ( col )
    {
      if ( is_false( col ) )
	c.marker_attr_def.doit = Marker_None;

      else
      {
	c.marker_attr_def.doit =
	  doit == Marker_ColString ? Marker_ColString :
	                             Marker_ColSymbol;
	
	strcpy( c.marker_col, col );
      }

      free( col );
    }

    /* if col is not set (i.e. no column name) and
       doit == Marker_ColString, then either there really was no
       column name, or there was an integer instead of alphanumeric
       stuff */
    else if ( Marker_ColString == doit )
      tf_exit( ExitERR_illegalArg,
	       "must specify a column name to use strings as a marker" );

    /* move spec pointer to beyond ':' so can process anything else */
    if ( dsetp )
      spec = dsetp + 1;
    else
      spec += strlen(spec);
  }

  c.do_markers = c.marker_attr_def.doit;

  /* is there a list of marker attributes, perchance?
     this'll return 1 for an empty string, so must do
     some extra checking */
  c.n_markers = str_tokqcnt( spec, ",", 0 );

  /* empty token means no markers */
  if ( c.n_markers == 1 && '\0' == *spec )
  {
    c.n_markers = 0;
  }

  else
  {
    char *ptr;
    char dchar;

    char *token;
    int i;

    MarkerAttr def_attr = c.marker_attr_def;

    c.marker_attr = (MarkerAttr *) malloc( c.n_markers * sizeof(MarkerAttr));

    if ( NULL == c.marker_attr )
      tf_exit( ExitERR_alloc, "unable to allocate %d markers", c.n_markers);

    for ( i = 0, token = str_tokq( spec, ",", 0, 1, &ptr, &dchar) ;
	  token && i < c.n_markers;
	  token = str_tokq( NULL, ",", 0, 1, &ptr, &dchar ), i++ )
    {
      c.marker_attr[i] = def_attr;

      def_attr.style++;

      if ( is_false( token ) )
	c.marker_attr[i].doit = Marker_None;

      else
      {
	char* res = parse_MarkAttr( token, NULL, &c.marker_attr[i] );

	if ( *res )
	  tf_message(
	    "ignoring extra extra attributes in marker attribute spec `%s'\n",
	    token );
      }

      /* update the global do_markers thing if we don't know already.
	 need this to allocate arrays in process() */
      if ( c.do_markers == Marker_None &&
	   c.marker_attr[i].doit != Marker_None )
	c.do_markers = c.marker_attr_def.doit;
    }

    tf_assert( i == c.n_markers && token == NULL );
  }

 leave:
  tf_leave();
}

static void
parse_connect( paramfile pf )
{
  char *dsetp;
  char dummy[STR_LENGTH+1];
  char *spec = dummy;

  tf_enter( "parse_connect" );

  tpgetstr( pf, "connect", spec, STR_LENGTH, 1 );

  strip_space( spec );

  if ( is_true( spec ) )
  {
    c.connect_attr_def.doit = Connect_Yes;
    goto leave;
  }
    
  if ( is_false( spec ) )
  {
    c.connect_attr_def.doit = Connect_None;
    goto leave;
  }


  if ( dsetp = strchr( spec, ':' ) )
  {
    char* res;

    *dsetp = '\0';

    res = parse_ConnectAttr( spec, &c.connect_attr_def );

    if ( *res )
      tf_message(
	"ignoring extra attributes in global connect attribute spec `%s'\n",
	dummy );

    /* move spec pointer to beyond ':' so can process anything else */
    spec = dsetp + 1;
  }
    

  /* is there a list of connect attributes, perchance?
     this'll return 1 for an empty string, so must do
     some extra checking */
  c.n_connect = str_tokqcnt( spec, ",", 0 );

  /* empty token means no connect */
  if ( c.n_connect == 1 && '\0' == *spec )
  {
    c.n_connect = 0;
  }

  else
  {
    char *ptr;
    char dchar;
    int i;
    char *token;

    ConnectAttr def_attr = c.connect_attr_def;

    c.connect_attr = (ConnectAttr *)
      malloc( c.n_connect * sizeof(ConnectAttr) );

    if ( NULL == c.connect_attr )
      tf_exit( ExitERR_alloc, "unable to allocate %d connect attributes",
	       c.n_connect);

    for ( i = 0, token = strtok( spec, "," ) ;
	  token && i < c.n_connect;
	  token = strtok( NULL, "," ), i++ )
    {
      char *res;
      c.connect_attr[i] = def_attr;

      def_attr.line.style = i % 5 + 1;

      res = parse_ConnectAttr( token, &c.connect_attr[i] );

      if (*res)
	tf_message(
	  "ignoring extra attributes in connect attribute spec `%s'\n",
	  token );

    }
    tf_assert( i == c.n_connect && token == NULL );
  }

  leave:
   tf_leave();

}
