#ifndef MAIN_H
#define MAIN_H

/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2007 Smithsonian Astrophysical Observatory
 *
 * This file is part of rdbplt
 *
 * rdbplt is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * rdbplt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stddef.h>
#include <stdio.h>

#include <suplib/str.h>

#define DEBUG_LENGTH 255
#define STR_LENGTH 255

#define PGPLOT_NSTYLES 5

enum
{
  ERR_NEG = 0, ERR_POS = 1
};


typedef struct
{
  float xmin, xmax, ymin, ymax;
} Box;

#define UNBOXP(box) &(box)->xmin,&(box)->xmax,&(box)->ymin,&(box)->ymax
#define UNBOX(box) (box)->xmin,(box)->xmax,(box)->ymin,(box)->ymax

/* the types of axis transformation supported */
typedef enum
{
  Axis_LINLIN,
  Axis_LINLOG,
  Axis_LOGLIN,
  Axis_LOGLOG
} AxisXFRM;

/* limit algorithms */
typedef enum
{
  LimitAlg_PGPLOT, LimitAlg_NONE, LimitAlg_PERCENT
} LimitAlg;

/* where the user has specified the legend */
typedef enum
{
  Legend_UL, Legend_UR, Legend_LL, Legend_LR
} LegendPos;

typedef struct
{
  double x, y;		/* lower left of legend box */
  char *text;		/* the various texts to write */
  char **textp;		/* pointers to the text */
  float char_ht;	/* character hieght */
  double lineskip;
  double line_length;
  double box_pad;
  LegendPos pos;
  int n;
  int box;
  int plot;		/* put a legend in the plot? */
} Legend;

typedef struct
{
  int style;
  int width;
  int color;
} LineAttr;

typedef enum
{
  Marker_None, Marker_ColString, Marker_ColSymbol, Marker_Symbol
} MarkerDoWhat;

typedef struct
{
  MarkerDoWhat doit;
  float ht;
  int width;
  int style;
  int color;
} MarkerAttr;

typedef enum
{
  Connect_None, Connect_Yes
} ConnectDoWhat;

typedef struct
{
  ConnectDoWhat doit;
  LineAttr line;
} ConnectAttr;

typedef struct
{
  int doit;
  float width;
  LineAttr line;
} ErrorBarAttr;

typedef struct
{
  /* pointers to the arrays which will hold the data */
  float* xdata;
  float* ydata;
  int*   markers;
  char** markers_str;
  float* xerr[2];
  float* yerr[2];

  ErrorBarAttr xerr_attr[2];
  ErrorBarAttr yerr_attr[2];
  MarkerAttr marker;
  ConnectAttr connect;

  char*  breakval;
  char* legend_text;
  
  int color;
  size_t nd;

} DataSet;


typedef struct
{
  char input [STR_LENGTH + 1];
  char device [STR_LENGTH + 1];
  char xlabel[STR_LENGTH + 1];
  char ylabel[STR_LENGTH + 1];
  char title [STR_LENGTH + 1];
  char subtitle [STR_LENGTH + 1];
  char xcol  [STR_LENGTH + 1];
  char ycol  [STR_LENGTH + 1];
  char xerr [2][STR_LENGTH + 1];
  char yerr [2][STR_LENGTH + 1];
  char marker_col  [STR_LENGTH + 1];
  char break_col [STR_LENGTH + 1];

  int flip_x;
  int flip_y;
  int do_log_x;
  int do_log_y;
  int do_break;

  double xmax, ymax, xmin, ymin;

  int auto_xmax, auto_xmin, auto_ymax, auto_ymin;
  int connect;
  int help;
  int version;
  int just;

  ConnectAttr   connect_attr_def;
  ConnectAttr  *connect_attr;
  int n_connect;

  MarkerAttr marker_attr_def;
  MarkerAttr *marker_attr;
  MarkerDoWhat do_markers;
  int n_markers;
  
  
  /* defaults */
  ErrorBarAttr xerr_attr_def[2];
  ErrorBarAttr yerr_attr_def[2];

  /* data set specific attributes */
  ErrorBarAttr *xerr_attr[2];
  int n_xerr_attr;

  ErrorBarAttr *yerr_attr[2];
  int n_yerr_attr;

  int n_colors;
  int *color;
  int color_cycle;
  int color_def;
  int linestyle_def;
  
  LimitAlg	limit_alg;
  double	limit_val;

  Legend legend;
  DataSet *dataset;
  int	n_datasets;

  double width, height;
  double image_pad;
  float char_height;
  int linewidth;

} Common;


/* gparam.c */
void gparam(int argc, char *argv[]);

/* init.c */
void set_def_attrs(void);
void init(void);

/* limits.c */
void get_limits(float *xmin, float *xmax, float *ymin, float *ymax);

/* main.c */
int main(int argc, char *argv[]);

/* plot.c */
void plot_data(float xmin, float xmax, float ymin, float ymax);

/* parse_attr.c */
char *strip_space(char *orig);
int is_true(char *str);
int is_false(char *str);
char *parse_LineAttr(char *desc, LineAttr *attr);
char *parse_ConnectAttr(char *desc, ConnectAttr *attr);
char *parse_ErrorBarAttr(char *desc, char **column, ErrorBarAttr *attr);
char *parse_MarkAttr(char *desc, char **col, MarkerAttr *attr);
void indent(FILE *out, int space);
void dump_LineAttr(FILE *out, LineAttr attr, int space);
void dump_ConnectAttr(FILE *out, ConnectAttr attr, int space);
void dump_MarkerAttr(FILE *out, MarkerAttr attr, int space);
void dump_ErrorBarAttr(FILE *out, ErrorBarAttr attr, char *which, int space);
void dump_Colors(FILE *out);
void dump_Attrs(FILE *out);
void dump_DataSetAttrs(FILE *out);

/* process.c */
void process(void);

#endif /* !MAIN_H */
