/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2007 Smithsonian Astrophysical Observatory
 *
 * This file is part of rdbplt
 *
 * rdbplt is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * rdbplt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

/* needed for strcasecmp on Solaris */
#ifdef SunOS
#  define __EXTENSIONS__
#endif
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <limits.h>
#include <stdio.h>

#include <suplib/str.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

/* remove any spaces */
char *
strip_space( char *orig )
{
  char* src = orig;
  char* dest = src;

  for ( ; *src ; src++ )
    if ( ! isspace( *src ) ) *dest++;
  *dest = '\0';

  return orig;
}

int
is_true( char *str )
{
  return 
    ( 0 == strcasecmp( "yes", str ) ||
      0 == strcasecmp( "true", str ) );
}

int is_false( char *str )
{
  return
    ( 0 == strcasecmp( "no", str ) ||
      0 == strcasecmp( "false", str ) ||
      0 == strcasecmp( "none", str ) ||
      0 == strcasecmp( "null", str ) );
}


#define DELIM "/"

/* line attributes have the format

   style/width/color

   it returns a pointer to the next character in the passed string
   after the last character in the attribute

*/
char *
parse_LineAttr(
  char *desc,	/* the attribute description to parse */

  LineAttr *attr  /* the attribute to update */
  )
{
  char *ptr;
  char dchar;

  char *token;

  tf_enter( "parse_LineAttr" );

  token = str_tokq( desc, DELIM, 0, 1, &ptr, &dchar );

  if ( token )
  {
    /* may be empty */
    if ( *token &&
	 -1 == (attr->style = tokmatch(token, &line_style_spec) ) )
    {
      /* restore string */
      str_tokq_restore( ptr, dchar );
      tf_exit( ExitERR_illegalArg, "unknown style: %s", desc );
    }

  }
  else
    goto leave;

  
  token = str_tokq( NULL, DELIM, 0, 1, &ptr, &dchar );
  
  if ( token )
  {
    /* may be empty */
    if ( *token )
    {
      char *p;
      attr->width = strtod( token, &p );

      if ( *p )
      {
	/* restore string */
	str_tokq_restore( ptr, dchar );
	tf_exit( ExitERR_illegalArg, "illegal line width: %s", desc );
      }
    }
  }
  else
    goto leave;


  token = str_tokq( NULL, DELIM, 0, 1, &ptr, &dchar );
  
  if ( token )
  {
    /* may be empty */
    if ( *token &&
	 -1 == ( attr->color = tokmatch(token, &color_spec) ) )
    {
      /* restore string */
      str_tokq_restore( ptr, dchar );
      tf_exit( ExitERR_illegalArg, "illegal line color: %s", desc );
    }
  }
  else
    goto leave;

  /* restore string */
  token = str_tokq_restore( ptr, dchar );

  tf_leave();
  return token ? token : desc + strlen(desc) ;

 leave:
  tf_leave();
  return desc + strlen(desc);


}


char *
parse_ConnectAttr(
  char *desc,	/* the attribute description to parse */

  ConnectAttr *attr  /* the attribute to update */
  )
{
  char *res = parse_LineAttr( desc, &attr->line );
  attr->doit = attr->line.style == 0 ? Connect_None : Connect_Yes;

  return res;
}

/* error bar attributes have the format

   [column/]crossbar_width/style/width/color

   it returns a pointer to the next character in the passed string
   after the last character in the attribute

*/
char *
parse_ErrorBarAttr(
  char *desc,	/* the attribute description to parse */
  char **column,	/* if non-null, expect a column name and
			   allocate space for it */
  ErrorBarAttr *attr  /* the attribute to update */
  )
{
  char *ptr;
  char *str = desc;
  char dchar;

  char *token;

  tf_enter( "parse_ErrorBarAttr" );

  if ( column )
  {
    token = str_tokq( str, DELIM, 0, 1, &ptr, &dchar );

    if ( token )
    {
      /* if empty token, complain */
      if (! *token)
	tf_exit( ExitERR_alloc, "missing column name" );

      if ( is_false( token ) )
      {
	attr->doit = 0;
	*column = NULL;
      }
      else
      {
	*column = str_dup( token );
	if ( NULL == *column )
	  tf_exit( ExitERR_alloc, "unable to dup column name" );

	attr->doit = 1;
      }
    }
    else
      goto leave;

    str = NULL;
  }

  token = str_tokq( str, DELIM, 0, 1, &ptr, &dchar );

  if ( token)
  {
    /* may be empty */
    if ( *token )
    {
      char *p;
      attr->width = strtod( token, &p );

      if ( *p )
      {
	/* restore string */
	token = str_tokq_restore( ptr, dchar );
	tf_exit( ExitERR_illegalArg, "illegal cross bar width: %s", desc );
      }
    }
  }
  else
    goto leave;

  if ( token = str_tokq_restore( ptr, dchar ) )
    ptr = parse_LineAttr( token, &attr->line );
  else
    goto leave;

  tf_leave();
  return ptr;

 leave:
  tf_leave();
  return desc + strlen(desc);

}


/* marker attributes have the format

   style/height/linewidth/color

   it returns a pointer to the next character in the passed string
   after the last character in the attribute

*/

char *
parse_MarkAttr(
  char *desc,	/* the attribute description to parse */
  char **col,
  MarkerAttr *attr  /* the attribute to update */
  )
{
  char *ptr;
  char dchar;

  char *token;

  tf_enter( "parse_MarkerAttr" );

  if ( is_false( desc ) )
  {
    attr->style = INT_MAX;
    goto leave;
  }

  token = str_tokq( desc, DELIM, 0, 1, &ptr, &dchar );

  if ( token )
  {
    /* may be empty */
    if ( *token )
    {
      char *p;
      attr->style = strtol( token, &p, 10 );

      if ( *p )
      {
	/* if it's not an integer, and interest was expressed
	   in a column name, save it */
	if ( col )
	{
	  *col = str_dup( token );
	  if ( NULL == *col )
	    tf_exit( ExitERR_alloc,
		     "unable to dup marker column name `%s'\n", token
		     );
	}

	/* nope, didn't ask for it */
	else
	{
	  str_tokq_restore( ptr, dchar );
	  tf_exit( ExitERR_illegalArg, "unknown marker style: %s", desc );
	}
      }

      /* this is now a given */
      else
	attr->doit = Marker_Symbol;

    }
  }
  else
    goto leave;

  
  token = str_tokq( NULL, DELIM, 0, 1, &ptr, &dchar );
  
  if ( token )
  {
    /* may be empty */
    if ( *token )
    {
      char *p;
      attr->ht = strtod( token, &p );

      if ( *p )
      {
	/* restore string */
	str_tokq_restore( ptr, dchar );
	tf_exit( ExitERR_illegalArg, "illegal marker height: %s", desc );
      }
    }
  }
  else
    goto leave;

  token = str_tokq( NULL, DELIM, 0, 1, &ptr, &dchar );
  
  if ( token )
  {
    /* may be empty */
    if ( *token )
    {
      char *p;
      attr->width = strtod( token, &p );

      if ( *p )
      {
	/* restore string */
	str_tokq_restore( ptr, dchar );
	tf_exit( ExitERR_illegalArg, "illegal marker line width: %s", desc );
      }
    }
  }
  else
    goto leave;


  token = str_tokq( NULL, DELIM, 0, 1, &ptr, &dchar );
  
  if ( token )
  {
    /* may be empty */
    if ( *token &&
	 -1 == ( attr->color = tokmatch(token, &color_spec) ) )
    {
      /* restore string */
      str_tokq_restore( ptr, dchar );
      tf_exit( ExitERR_illegalArg, "illegal marker color: %s", desc );
    }
  }
  else
    goto leave;

  /* restore string */
  token = str_tokq_restore( ptr, dchar );

  tf_leave();
  return token ? token : desc + strlen(desc) ;

 leave:
  tf_leave();
  return desc + strlen(desc);


}

void
indent( FILE *out, int space )
{
  int i = 0;
  while ( i++ < space )
    fputc( ' ', out );
}

void
dump_LineAttr( FILE *out, LineAttr attr, int space )
{
  indent(out,space); fprintf( out, "line:\n" );
  space += 2;
  indent(out,space); fprintf( out, "style = %d\n", attr.style );
  indent(out,space); fprintf( out, "width = %d\n", attr.width );
  indent(out,space); fprintf( out, "color = %d\n", attr.color );
}

void
dump_ConnectAttr( FILE *out, ConnectAttr attr, int space )
{
  indent(out,space); fprintf( out, "connect:\n" );
  space += 2;
  indent(out,space); fprintf( out, "doit = %d\n", attr.doit );
  dump_LineAttr(out, attr.line, space);
}

void
dump_MarkerAttr( FILE *out, MarkerAttr attr, int space )
{
  indent(out,space); fprintf( out, "marker:\n" );
  space += 2;
  indent(out,space); fprintf( out, "doit = %d\n", attr.doit );
  indent(out,space); fprintf( out, "ht = %f\n", attr.ht );
  indent(out,space); fprintf( out, "style = %d\n", attr.style );
  indent(out,space); fprintf( out, "width = %d\n", attr.width );
  indent(out,space); fprintf( out, "color = %d\n", attr.color );
}

void
dump_ErrorBarAttr( FILE *out, ErrorBarAttr attr, char* which,int space )
{
  indent(out,space); fprintf( out, "%s error bar:\n", which );
  space += 2;
  indent(out,space); fprintf( out, "doit = %d\n", attr.doit );
  indent(out,space); fprintf( out, "width = %f\n", attr.width );
  dump_LineAttr(out, attr.line, space);
}

void
dump_Colors( FILE *out )
{
  if ( c.n_colors )
  {
    int i;
    fprintf( out, "\nColors:\n" );
    for ( i = 0 ; i < c.n_colors; i++ )
      fprintf( out, "%d: %d\n", i, c.color[i] );
  }
}


void
dump_Attrs( FILE *out )
{
  fprintf( out, "\nconnect_attr_def\n" );
  dump_ConnectAttr( out, c.connect_attr_def, 2 );

  fprintf( out, "\nmarker_attr_def\n" );
  dump_MarkerAttr( out, c.marker_attr_def, 2 );

  if ( c.n_connect )
  {
    int i;
    for ( i = 0; i < c.n_connect ; i++)
    {
      fprintf( out, "\n  %2d:\n", i );
      dump_ConnectAttr( out, c.connect_attr[i], 2 );
    }
  }

  if ( c.n_markers )
  {
    int i;
    for ( i = 0; i < c.n_markers ; i++)
    {
      fprintf( out, "\n  %2d:\n", i );
      dump_MarkerAttr( out, c.marker_attr[i], 2 );
    }
  }

}

void
dump_DataSetAttrs( FILE *out )
{
  int i;

  for ( i = 0 ; i < c.n_datasets ; i++ )
  {
    fprintf( out, "\ndata set %d\n", i );
    dump_MarkerAttr(out, c.dataset[i].marker, 2 );
    dump_ConnectAttr(out, c.dataset[i].connect, 2 );
    dump_ErrorBarAttr(out, c.dataset[i].xerr_attr[0], "xerr n", 2 );
    dump_ErrorBarAttr(out, c.dataset[i].xerr_attr[1], "xerr p", 2 );
    dump_ErrorBarAttr(out, c.dataset[i].yerr_attr[0], "yerr n", 2 );
    dump_ErrorBarAttr(out, c.dataset[i].yerr_attr[1], "yerr p", 2 );
    
  }
}
