/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2007 Smithsonian Astrophysical Observatory
 *
 * This file is part of rdbplt
 *
 * rdbplt is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * rdbplt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <string.h>

#include <cpgplot.h>

#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>

#include "main.h"
#include "common.h"

static void	legend( void );
static void	setup_viewport(float xmin, float xmax,
			       float ymin, float ymax);

static void	clean_err_bars(float min, float max, int direction,
			       float data[], float err[], float n );

static void
set_LineAttr ( LineAttr attr )
{
  cpgsls( attr.style );
  cpgslw( attr.width );
  cpgsci( attr.color );
}

static void
set_ConnectAttr ( ConnectAttr attr )
{
  set_LineAttr( attr.line );
}

static void
set_MarkerAttr ( MarkerAttr attr )
{
  cpgsch( attr.ht );
  cpgslw( attr.width );
  cpgsci( attr.color );
}

static void
set_ErrorBarAttr ( ErrorBarAttr attr )
{
  set_LineAttr( attr.line );
  cpgsch( attr.width );
}

void
plot_data(
  float xmin,
  float xmax,
  float ymin,
  float ymax
  )
{
  int i;
  int errb;

  /* start up PGPLOT, and set up the frame, axis labels, and title */
  if ( cpgopen( c.device ) < 0 )
    return;

  cpgsch( c.char_height );
  cpgslw( c.linewidth );

  get_limits( &xmin, &xmax, &ymin, &ymax );
  
  setup_viewport( xmin, xmax, ymin, ymax );

  cpglab( c.xlabel, c.ylabel, c.title );
  if ( strlen( c.subtitle ) )
    cpgmtxt( "T", 0.5, 0.5, 0.5, c.subtitle );

  for ( i = 0 ; i < c.n_datasets ; i++ )
  {
    /* plot the error bars */
    for( errb = 0 ; errb < 2 ; errb++ )
    {
      /* direction is -1 for i=0, 1 for i=1 */
      int direction = ERR_POS == errb ? 1 : -1;

      if ( c.dataset[i].xerr_attr[errb].doit )
      {
	cpgsave();
	set_ErrorBarAttr( c.dataset[i].xerr_attr[errb] );

	/* PGPLOT can't deal with error bars going to -FLT_MAX,
	   so set them to the edge of the viewport */
	clean_err_bars( xmin, xmax, direction,
			c.dataset[i].xdata,
			c.dataset[i].xerr[errb],
			c.dataset[i].nd );

	/* plot direction is 1 for x+e, 3 for x-e */
	cpgerrb( 2 - direction, c.dataset[i].nd,
		 c.dataset[i].xdata,
		 c.dataset[i].ydata,
		 c.dataset[i].xerr[errb],
		 1.0 );
	cpgunsa();
      }

      if ( c.dataset[i].yerr_attr[errb].doit )
      {
	cpgsave();
	set_ErrorBarAttr( c.dataset[i].yerr_attr[errb] );

	/* PGPLOT can't deal with error bars going to -FLT_MAX,
	   so set them to the edge of the viewport */
	clean_err_bars( ymin, ymax, direction,
			c.dataset[i].ydata,
			c.dataset[i].yerr[errb],
			c.dataset[i].nd );

	/* plot direction is 2 for y+e, 4 for y-e */
	cpgerrb( 3 - direction, c.dataset[i].nd,
		 c.dataset[i].xdata,
		 c.dataset[i].ydata,
		 c.dataset[i].yerr[errb],
		 1.0 );
	cpgunsa();
      }
    }

    /* if markers are strings from a column in the table */
    if ( Marker_ColString == c.dataset[i].marker.doit )
    {
      int pt;
      cpgsave();
      set_MarkerAttr( c.dataset[i].marker );

      for ( pt = 0 ; pt < c.dataset[i].nd ; pt++ )
      {
	cpgtext( c.dataset[i].xdata[pt],
		 c.dataset[i].ydata[pt],
		 c.dataset[i].markers_str[pt] );
      }
      cpgunsa();
    }

    /* if markers are styles from a column in the table */
    else if ( Marker_ColSymbol == c.dataset[i].marker.doit )
    {
      cpgsave();
      set_MarkerAttr( c.dataset[i].marker );

      cpgpnts( c.dataset[i].nd,
	       c.dataset[i].xdata,
	       c.dataset[i].ydata,
	       c.dataset[i].markers,
	       c.dataset[i].nd );
      cpgunsa();
    }

    /* if markers were specified via the marker parameter and this
       one is on */
    else if ( Marker_Symbol == c.dataset[i].marker.doit )
    {
      cpgsave();
      set_MarkerAttr( c.dataset[i].marker );

      cpgpt( c.dataset[i].nd,
	     c.dataset[i].xdata,
	     c.dataset[i].ydata,
	     c.dataset[i].marker.style );
      cpgunsa();
    }

    if ( Connect_Yes == c.dataset[i].connect.doit )
    {
      cpgsave();
      set_LineAttr( c.dataset[i].connect.line );

      cpgline( c.dataset[i].nd,
	       c.dataset[i].xdata,
	       c.dataset[i].ydata );
      cpgunsa();
    }

  }

  if ( c.legend.plot && c.legend.n > 0)
    legend( );
  
  cpgclos( );
}




static void
legend( void )
{
  Box vwsurf, vp, bounds;
  int i;
  char string[1024];

  float leg_x, leg_y, line_length;
  float x, y, xch, ych;
  float max_width = -FLT_MAX;
  float xoffset = 0;
  int ls;
  float char_ht;

  tf_enter( "legend" );

  /* get size of viewsurface, in inches */
  cpgqvsz( 1, UNBOXP(&vwsurf) );

  /* get stuff so can restore, later */
  cpgqvp( 1, UNBOXP(&vp ) );
  cpgqwin( UNBOXP( &bounds ) );
  cpgqls( &ls );
  cpgqch( &char_ht );

  /* grab entire viewsurface as viewport */
  cpgvsiz( UNBOX(&vwsurf) );
  cpgswin( UNBOX(&vwsurf) );

  cpgsch( c.legend.char_ht );
  cpgqcs( 1, &xch, &ych );
  cpgsls( 1 );

  /* determine position of legend box in viewsurface coords (in inches) */
  leg_x = c.legend.x * ( vp.xmax - vp.xmin ) + vp.xmin;
  leg_y = c.legend.y * ( vp.ymax - vp.ymin ) + vp.ymin;
  line_length = c.legend.line_length * ( vp.xmax - vp.xmin );

  switch( c.legend.pos )
  {
  case Legend_UL: break;

  case Legend_LL:
    leg_y += c.n_datasets * ych * c.legend.lineskip;
    if ( c.legend.box )
      leg_y += 2 * c.legend.box_pad * ych;
    break;

  default:
    tf_assert( 0 );
  }


  x = leg_x;
  y = leg_y - ych;

  if ( c.legend.box )
  {
    x += c.legend.box_pad * xch;
    y -= c.legend.box_pad * ych;
  }

  for ( i = 0 ; i < c.n_datasets ; i++ )
  {
    float xbox[4], ybox[4];

    if ( c.connect )
      xoffset = line_length + xch;
    else if  ( c.dataset[i].marker.doit != Marker_None )
      xoffset = xch;

    cpgsave();

    cpgsci( c.dataset[i].color );

    if ( Connect_Yes == c.dataset[i].connect.doit )
    {
      cpgsave();
      set_LineAttr( c.dataset[i].connect.line );

      cpgmove( x, y + 0.25 * ych );
      cpgdraw( x + line_length, y + 0.25 * ych );
      cpgunsa();
    }

    if ( c.dataset[i].marker.doit != Marker_None )
    {
      float xt = c.connect ? x + ( xoffset - xch ) / 2.0 : x ;
      float yt = y + 0.25 * ych;

      cpgsave();
      set_MarkerAttr( c.dataset[i].marker );

      cpgpt( 1, &xt, &yt, c.dataset[i].marker.style );
      cpgunsa();
    }

    cpgsch( c.legend.char_ht );

    if ( c.dataset[i].breakval )
      sprintf( string, c.dataset[i].legend_text, c.dataset[i].breakval );
    else
      strcpy( string, c.dataset[i].legend_text );
    cpgtext( x + xoffset, y, string );

    cpgqtxt( x + xoffset, y, 0, 0, string, (float *) &xbox, (float *) &ybox );
    if ( xbox[3] - xbox[0] > max_width )
      max_width = xbox[3] - xbox[0];

    y -= c.legend.lineskip * ych;

    cpgunsa();
  }

  if ( c.legend.box )
  {
    double width = max_width + 2 * c.legend.box_pad * xch + xoffset;
    double height =  2 * c.legend.box_pad * ych +
         c.n_datasets * ych * c.legend.lineskip;

    cpgsls( 1 );

    cpgmove( leg_x        , leg_y );
    cpgdraw( leg_x + width, leg_y );
    cpgdraw( leg_x + width, leg_y - height );
    cpgdraw( leg_x        , leg_y - height );
    cpgdraw( leg_x        , leg_y );
  }

  cpgsch( char_ht );
  cpgsls( ls );
  cpgvsiz( UNBOX(&vp) );
  cpgswin( UNBOX(&bounds) );
}


#define BOX_OPTS_LINEAR "BCNTS"
#define BOX_OPTS_LOG	"BCNTSL"
static void
setup_viewport( 
  float xmin,
  float xmax,
  float ymin,
  float ymax
  )
{
  Box vwsurf, vp;
  double width, height;
  float xch, ych;

  /* get size of viewsurface, in inches */
  cpgqvsz(1, UNBOXP(&vwsurf));

  /* character height, in inches */
  cpgqcs( 1, &xch, &ych );

  /* transform into viewport size */
  width  = vwsurf.xmax - vwsurf.xmin - 2 * c.image_pad * xch;
  height = vwsurf.ymax - vwsurf.ymin - 2 * c.image_pad * ych;

  if ( c.width )
    width = c.width;

  if ( c.height )
    height = c.height;

  /* if axis scaling is requested, alter viewport size */
  if ( c.just )
  {
    double aspect = fabs( ( ymax - ymin ) / ( xmax - xmin ) );

    if (aspect > height / width)
      width = height / aspect;
    else
      height = width * aspect;
  }

  /* center viewport in viewsurface */
  vp.xmin = ( vwsurf.xmax + vwsurf.xmin - width ) / 2.0 ;
  vp.xmax = vp.xmin + width;

  vp.ymin = ( vwsurf.ymax + vwsurf.ymin - height ) / 2.0 ;
  vp.ymax = vp.ymin + height;

  /* now set new viewport */
  cpgvsiz( UNBOX(&vp) );

  if ( c.flip_x )
  {
    float tmp = xmin;
    xmin = xmax;
    xmax = tmp;
  }

  if ( c.flip_y )
  {
    float tmp = ymin;
    ymin = ymax;
    ymax = tmp;
  }

  cpgswin( xmin, xmax, ymin, ymax );
  cpgbox( c.do_log_x ? BOX_OPTS_LOG : BOX_OPTS_LINEAR, 0, 0,
	  c.do_log_y ? BOX_OPTS_LOG : BOX_OPTS_LINEAR, 0, 0 );

}


/* PGPLOT can't deal with error bars going to -FLT_MAX,
   so set them to just beyond the edge of the viewport.
*/

static void
clean_err_bars(
  float min,
  float max,
  int   direction,
  float data[],
  float err[],
  float n )
{
  int i;

  for ( i = 0 ; i < n ; i++, err++, data++ )
  {
    double res = *data + direction * *err;
    if ( res < min ) *err = min - res + 1;
    if ( res > max ) *err = res - max + 1;
  }
}
