/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2007 Smithsonian Astrophysical Observatory
 *
 * This file is part of rdbplt
 *
 * rdbplt is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * rdbplt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include "config.h"

#ifdef HAVE_IEEEFP
#include <ieeefp.h>
#endif

#include <stdio.h>
#include <float.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <limits.h>

#include <mst_rdb/mst_rdb.h>		/* for rdb reading routines */

#include <suplib/str.h>
#include <suplib/debug.h>
#include <tracefct/tracefct.h>
#include <tracefct/exiterrvals.h>


#include "main.h"
#include "common.h"

/* the rdb routines need a structure to fill with data.  this declares it */
typedef struct
{
  double x;
  double y;
  double xerr[2];
  double yerr[2];
  double marker;
  char * marker_str;
  char * break_col;
} Data;
#define SETFIELD(col,var,vartype,idx)		\
do{						\
    fields[idx].name = check_col( hdr, col );	\
    fields[idx].type = vartype;			\
    fields[idx].offset = offsetof( Data, var );	\
    idx++;       				\
}while(0)
	

typedef struct
{
  void *data;
  size_t size;
  int is_string;
} Buffer;

#define SETBUF(array,idx) \
do{						\
	buffer[idx].size = sizeof( *array ) ;	\
	buffer[idx].data = &array; 		\
	buffer[idx].is_string = 0;	 	\
	idx++;					\
}while(0)

#define SETBUFSTR(array,idx) \
do{						\
	buffer[idx].size = sizeof( *array ) ;	\
	buffer[idx].data = &array; 		\
	buffer[idx].is_string = 1; 		\
	idx++;					\
}while(0)

/* forward references */
static void	alloc_data( Buffer buffer[], size_t nbufs, size_t *buflen );
static void	free_data( Buffer buffer[], size_t nbufs, size_t nd );
static void	resize_data( Buffer buffer[], size_t nbufs, size_t buflen );
static char *	check_col( rdbHeader *, char *col );
static void	setup_dataset( DataSet* dset, int idx, int* marker_style,
			       int* line_style, int* color );
static void	alloc_dataset( void );
static void	free_datasets( void );
static void	cleanup_dataset( DataSet *dset, float *xdata,
				float *ydata, int *markers, char** markers_str,
				float *xerr[2], float *yerr[2], int nd);

static void	error_bar(double data, double tdata, double error,
			  int direction, int do_log, int auto_min,
			  int auto_max, float *error_bar,
			  float *min, float *max);

static void	preset_dset_attr( DataSet* dset );

/* see if the current x limits need to be changed, based upon a given value */
#define TEST_X(val) \
do {\
      float __val = val; \
      if ( c.auto_xmin && __val < xmin ) xmin = __val; \
      if ( c.auto_xmax && __val > xmax ) xmax = __val; \
}while (0)
	
/* see if the current y limits need to be changed, based upon a given value */
#define TEST_Y(val) \
do {\
      float __val = val; \
      if ( c.auto_ymin && __val < ymin ) ymin = __val; \
      if ( c.auto_ymax && __val > ymax ) ymax = __val; \
}while (0)

/*
   This function does most of the work.  It opens the rdb file, reads
   the data, transforming them if necessary, then plots them.

   The data are stored in several arrays of floats (PGPLOT only does
   floats).  The data for the min and max of the error bars are
   stored separately, in part because PGPLOT needs them that way,
   but more importantly because the min and max won't be the same
   if the data are transformed via logarithms.

   The arrays for the data are dynamically increased as necessary,
   allowing one to read the data from the standard input.  Since
   the number of data arrays that need to be allocated changes if
   there are errorbars, the allocation function operates on an
   array of pointers to the data arrays.

 */

void
process( void )
{
  FILE *fin;

  int i;
  int nfields;			/* the number of columns to read */
  int idx = 0;			/* which dataset is being read in */

  /* pointers to the arrays which will hold the data */
  float* xdata = NULL;
  float* ydata = NULL;
  int*   markers = NULL;
  char** markers_str = NULL;
  float* xerr[2] = { NULL, NULL };
  float* yerr[2] = { NULL, NULL };

  /* tmp pointers to the above arrays, for speed */
  float* p_xdata = NULL;
  float* p_ydata = NULL;
  int*   p_markers = NULL;
  char** p_markers_str = NULL;
  float* p_xerr[2] = { NULL, NULL };
  float* p_yerr[2] = { NULL, NULL };

  /* values of the axes limits, in input data space */
  float xmax_lim, xmin_lim;
  float ymax_lim, ymin_lim;

  int line_style = c.connect_attr_def.line.style;
  int marker_style = c.marker_attr_def.style;
  int color = c.color_def;

  Buffer buffer[128];	/* pointers to the arrays which are actually used */
  
  size_t nbufs = 0;	/* the number of arrays used */
  size_t buflen = 0;	/* the number of elements in a buffer (not chars) */

  Data data;		/* where the rdb routine stores its data */

  size_t nr, nd;
  float xmin, xmax, ymin, ymax;		/* the limits of the data */

  int do_x_NR = 0;	/* if true, x value is row index */
  int do_y_NR = 0;	/* if true, y value is row index */

  /* various rdb things */
  rdbHeader *hdr;
  DataColumnMap_st *map;
  RDBFieldStInfo fields[10];	/* will only ever need 6 */

  tf_enter( "process" );

  /* open up the rdb file, and read its header */
  if ( strcmp( c.input, "stdin" ) )
  {
    if ( NULL == ( fin = fopen( c.input, "r" ) ) )
      tf_exit( ExitERR_fopen, "couldn't open %s", c.input );
  }
  else
    fin = stdin;

  if ( NULL == (hdr = rdb_rd_hdr( fin ) ) )
    tf_exit( ExitERR_misc, "nothing in rdb file?" );

  /*
    the rdb routines need to know which columns we're interested in.
    since we know this only at compile time, build the map array dynamically.
    note that check_col is used to determine if the user specified a
    column by name or by index.  note also that the column may have the
    special name "_NR", which indicates it should be the row index
  */

  nfields = 0;

  if ( strcmp( c.ycol, "_NR" ) )
    SETFIELD( c.ycol, y, RDB_Num, nfields );
  else
    do_y_NR++;		/* y will be the row index.  generated below */

  SETBUF(ydata,nbufs);
  
  if ( strcmp( c.xcol, "_NR" ) )
    SETFIELD( c.xcol, x, RDB_Num, nfields );
  else
    do_x_NR++;		/* x will be the row index.  generated below */

  SETBUF(xdata,nbufs);
  
  /* if a column was specified for the x error, do x error bars */
  for ( i = 0 ; i < 2 ; i++ )
  {
    if ( c.xerr_attr_def[i].doit > 0 )
    {
      fields[nfields].name = check_col( hdr, c.xerr[i] );
      fields[nfields].type = RDB_Num;
      fields[nfields].offset = offsetof( Data, xerr ) + i * sizeof(double);
      nfields++;

      SETBUF(xerr[i],nbufs);
    }

    /* if a column was specified for the y error, do y error bars */
    if ( c.yerr_attr_def[i].doit > 0 )
    {
      fields[nfields].name = check_col( hdr, c.yerr[i] );
      fields[nfields].type = RDB_Num;
      fields[nfields].offset = offsetof( Data, yerr ) + i * sizeof(double);
      nfields++;
      SETBUF(yerr[i],nbufs);
    }
  }
  /* if a marker column was specified, use it, but be careful about
     strings versus marker style
  */
  if ( Marker_ColString == c.do_markers )
  {
    SETFIELD( c.marker_col, marker_str, RDB_String, nfields );
    SETBUFSTR(markers_str,nbufs);
  }
  else if ( Marker_ColSymbol == c.do_markers )
  {
    SETFIELD( c.marker_col, marker, RDB_Num, nfields );
    SETBUF(markers,nbufs);
  }

  /* if a break column was specified, read it in as a string,
     regardless of what it is */
  if ( c.do_break > 0 )
  {
    SETFIELD( c.break_col, break_col, RDB_String, nfields );
  }


  /* generate a mapping between what we want and what's in the file */
  map = rdb_map_cols_stst( hdr, nfields, fields );

  /* set the limits based upon which axes the user wants auto scaled */
  xmax = c.auto_xmax ? -FLT_MAX : c.xmax;
  xmin = c.auto_xmin ?  FLT_MAX : c.xmin;
  ymax = c.auto_ymax ? -FLT_MAX : c.ymax;
  ymin = c.auto_ymin ?  FLT_MAX : c.ymin;

  /* set the limit comparison values. these must be in input (pre-xfrm)
     units */
  if ( ! c.auto_xmax )
    xmax_lim = c.do_log_x ? pow( 10, c.xmax ) : c.xmax;
  if ( ! c.auto_xmin )
    xmin_lim = c.do_log_x ? pow( 10, c.xmin ) : c.xmin;

  if ( ! c.auto_ymax )
    ymax_lim = c.do_log_y ? pow( 10, c.ymax ) : c.ymax;
  if ( ! c.auto_ymin )
    ymin_lim = c.do_log_y ? pow( 10, c.ymin ) : c.ymin;

  /* use this to determine if a break column was specified, if it
     was, this will be non-NULL */
  data.break_col = NULL;

  c.n_datasets = 0;
  alloc_dataset();

  /* presets, in case there is only one dataset */
  setup_dataset( &c.dataset[0], 0, &marker_style, &line_style, &color );

  /* read the file, processing the data as we go */
  for ( idx = nd = 0, nr = 1 ; rdb_col_read_st( fin, hdr, map, &data ) ; nr++)
  {
    /* check data; do the mandatory columns first */
    if ( ! finite(data.x) )
    {
      tf_message( "column %s: bad data value at row %d\n",
		  c.xcol, nr );
      continue;
    }

    if ( ! finite(data.y) )
    {
      tf_message( "column %s: bad data value at row %d\n",
		  c.ycol, nr );
      continue;
    }

    /* if we're looking for a break column, check it */
    if ( data.break_col )
    {
      /* check if first element in first data set */
      if ( NULL == c.dataset[idx].breakval )
      {
	if ( NULL ==
	     ( c.dataset[idx].breakval = str_dup(data.break_col ) ) )
	  tf_exit( ExitERR_alloc, "couldn't dup breakval" );
      }

      /* previous break val isn't the same as the current one */
      else if ( strcmp( c.dataset[idx].breakval, data.break_col ) )
      {
	int ii;

	resize_data( buffer, nbufs, nd );
	/* need to do this after a dataset is read in to track
	   the data pointers */
	cleanup_dataset( &c.dataset[idx],
			 xdata, ydata, markers, markers_str,
			 xerr, yerr, nd);
	alloc_dataset( );
	idx++;
	buflen = 0;
	nd = 0;

	setup_dataset( &c.dataset[idx], idx, &marker_style, &line_style, &color );

	if ( NULL ==
	     ( c.dataset[idx].breakval = str_dup(data.break_col ) ) )
	  tf_exit( ExitERR_alloc, "couldn't dup breakval" );

      }

      free( data.break_col );
    }

    /* if the data buffers are full, re-allocate them */
    if ( nd == buflen )
    {
      alloc_data( buffer, nbufs, &buflen );

      p_xdata    = &xdata[nd];
      p_ydata    = &ydata[nd];

      /* note that this is legal even if markers and markers_str
	 were never allocated; they'll be garbage, but they're
	 never dereferenced */ 
      p_markers  = &markers[nd];
      p_markers_str  = &markers_str[nd];

      for ( i = 0 ; i < 2 ; i++ )
      {
	p_xerr[i] = &xerr[i][nd];
	p_yerr[i] = &yerr[i][nd];
      }
    }

    /* if the x data are to be row indices, make them */
    if ( do_x_NR )
      data.x = nr;

    if ( do_y_NR )
      data.y = nr;

    /* make sure we don't handle data that isn't supposed to be
       plotted.  it screws up the autoscale stuff */
    if ( (!c.auto_xmax && data.x > xmax_lim) ||
	 (!c.auto_xmin && data.x < xmin_lim) ||
	 (!c.auto_ymax && data.y > ymax_lim) ||
	 (!c.auto_ymin && data.y < ymin_lim) )
      continue;

    /* now transform the data */
    if ( c.do_log_x )
    {
      if ( data.x > 0 )
      {
	*p_xdata = log10( data.x );
	TEST_X( *p_xdata );
      }
      else
      {
	tf_message( "column %s: row %d: log of non-positive "
		    "value - skipping\n",
		      c.xcol, nr );
	continue;
      }
    }
    else
    {
      *p_xdata = data.x;
      TEST_X( data.x );
    }

    if ( c.do_log_y )
    {
      if ( data.y > 0 )
      {
	*p_ydata = log10( data.y );
	TEST_Y( *p_ydata );
      }
      else
      {
	tf_message( "column %s: row %d: log of non-positive "
		    "value - skipping\n",
		      c.ycol, nr );
	continue;
      }
    }
    else
    {
      *p_ydata = data.y;
      TEST_Y( data.y );
    }


    /* in an attempt to be time efficient, this section duplicates lots
       of code.  error bar limits are calculated (before the data are
       transformed) and checked against plot limits.  to make the
       plot useful for log transforms, non-positive pre-transform
       data are not included in plot limit checks.
    */
    for ( i = 0 ; i < 2 ; i++ )
    {
      /* direction is -1 for i=0, 1 for i=1 */
      int direction = ERR_POS == i ? 1 : -1;

      if ( c.dataset[idx].xerr_attr[i].doit )
      {
	if (! finite( data.xerr[i] ) )
	{
	  tf_message( "column %s: bad data value at row %d\n",
		      c.xerr[i], nr );
	  continue;
	}
	error_bar( data.x, *p_xdata, data.xerr[i], direction,
		   c.do_log_x, c.auto_xmin, c.auto_xmax,
		   p_xerr[i], &xmin, &xmax );
      }


      if ( c.dataset[idx].yerr_attr[i].doit )
      {
	if ( ! finite( data.yerr[i] ) )
	{
	  tf_message( "column %s: bad data value at row %d\n",
		      c.yerr[i], nr );
	  continue;
	}

	error_bar( data.y, *p_ydata, data.yerr[i], direction,
		   c.do_log_y, c.auto_ymin, c.auto_ymax,
		   p_yerr[i], &ymin, &ymax );
      }
    }

    /* the above loop will be prematurely aborted upon error; check
       for that and continue with the next data point */
    if ( i != 2 )
      continue;

    /* if markers are to be read from the file, save the current one */
    if ( markers_str )
      *p_markers_str = data.marker_str;
      
    else if ( markers )
      *p_markers = data.marker;

    /* increment number of data points.  do this after all the work
       is done because array indexing is zero-based (duh!)
    */
    nd++;

    p_xdata++;
    p_ydata++;
    p_markers++;
    p_markers_str++;
    p_xerr[0]++;
    p_xerr[1]++;
    p_yerr[0]++;
    p_yerr[1]++;
  }

  /* clean up last dataset */
  resize_data( buffer, nbufs, nd );
  cleanup_dataset( &c.dataset[idx],
		   xdata, ydata, markers, markers_str,
		   xerr, yerr, nd);

  if ( FLT_MAX == xmin || -FLT_MAX == xmax )
  {
    xmin = -1;
    xmax = - xmin;
  }

  if ( FLT_MAX == ymin || -FLT_MAX == ymax )
  {
    ymin = -1;
    ymax = - ymin;
  }

  if ( dbflag( "dump" ) )
  {
    dump_Attrs( stderr );
    dump_DataSetAttrs( stderr );
  }
  plot_data( xmin, xmax, ymin, ymax );

  free_data( buffer, nbufs, nd );

  rdb_free_map( map );

  rdb_free_hdr( hdr );
}

#define DATA_SIZE 8192

/*
  (re) allocate a bunch of arrays.  they all are of the same type
   and have the same length
*/
static void
alloc_data(
  Buffer buffer[],	/*  pointers to the variables which hold
			    pointers to the arrays (so that they can
			    be changed upon a reallocation) */
  size_t nbufs,		/*  the number of arrays */
  size_t *buflen	/*  a pointer to a variable holding the
			    current size of the arrays.  this variable
			    should be set to zero for the first call */
)
{
  int i;
  int first_alloc = 0 == *buflen;
  
  *buflen += DATA_SIZE;

  for ( i = 0 ; i < nbufs ; i++ )
  {
    void **data = buffer[i].data;
    size_t size = buffer[i].size;

    if ( first_alloc )
    {
      if ( NULL == (*data = malloc( *buflen * size )) )
	tf_exit( ExitERR_alloc, "couldn't allocate more data space" );
    }
    else
    {
      if ( NULL == (*data = realloc( *data, *buflen * size )) )
	tf_exit( ExitERR_alloc, "couldn't allocate more data space" );
    }
  }
}

/* free the data allocated by alloc_data */
static void
resize_data( Buffer buffer[], size_t nbufs, size_t buflen )
{
  int i;
  
  for ( i = 0 ; i < nbufs ; i++ )
  {
    void **data = buffer[i].data;
    size_t size = buffer[i].size;

    if ( NULL == (*data = realloc( *data, buflen * size )) )
	tf_exit( ExitERR_alloc, "couldn't resize" );
  }

}

/* free the data allocated by alloc_data */
static void
free_data( Buffer buffer[], size_t nbufs, size_t nd )
{
  int i;
  
  for ( i = 0 ; i < nbufs ; i++ )
    if ( NULL != *((void **) buffer[i].data) )
    {

      if ( buffer[i].is_string )
      {
	int j;
	char **ptr = *((char ***) buffer[i].data);
	for ( j = 0 ; j < nd ; j++ )
	{
	  free( ptr[j] );
	}
      }

      free( *((void **) buffer[i].data) );
    }
}

/*
  check if a column name is really a column name, or if it's a column
  index, which is designated by '_N' where N is an unsigned integer.
*/
static char *
check_col( rdbHeader *hdr, char *col )
{
  char *rcol = col;

  tf_enter( "check_col" );

  if ( '_' == col[0] )
  {
    char *ptr;
    unsigned long nc = strtoul( col+1, &ptr, 10 );

    if ( *ptr && ! isspace(*ptr) )
      tf_exit(ExitERR_illegalArg, "illegalcolumn name: %s", col );

    if ( nc == 0 || nc > hdr->n )
      tf_exit(ExitERR_illegalArg, "column number out of range: %s", col );

    rcol = hdr->col_name[nc-1];
  }

  tf_leave( );
  return rcol;
}

static void
alloc_dataset( void )
{
  int i;
  c.n_datasets++;

  if ( 0 == c.n_datasets )
  {
    if ( NULL == (c.dataset = malloc( sizeof( DataSet ) )) )
      tf_exit( ExitERR_alloc, "unable to allocate dataset" );
  }
  else
  {
    if ( NULL ==
	 (c.dataset = realloc( c.dataset, c.n_datasets * sizeof(DataSet))) )
      tf_exit( ExitERR_alloc, "unable to allocate dataset" );
  }

  c.dataset[c.n_datasets-1].xdata    = NULL;
  c.dataset[c.n_datasets-1].ydata    = NULL;
  c.dataset[c.n_datasets-1].markers  = NULL;

  for ( i = 0 ; i < 2 ; i++ )
  {
    c.dataset[c.n_datasets-1].xerr[i] = NULL;
    c.dataset[c.n_datasets-1].yerr[i] = NULL;
  }
  c.dataset[c.n_datasets-1].breakval = NULL;
  c.dataset[c.n_datasets-1].nd       = 0;
}

static void
cleanup_dataset(
  DataSet* dset,
  float* xdata,
  float* ydata,
  int* markers,
  char** markers_str,
  float* xerr[2],
  float* yerr[2],
  int nd
  )
{
  int i;

  dset->xdata    = xdata;
  dset->ydata    = ydata;
  dset->markers  = markers;
  dset->markers_str  = markers_str;
  dset->nd       = nd;

  for ( i = 0 ; i < 2 ; i++ )
  {
    dset->xerr[i] = xerr[i];
    dset->yerr[i] = yerr[i];
  }

}

static void
free_datasets( void )
{
  int i;

  if ( NULL == c.dataset )
    return;

  for ( i = 0 ; i < c.n_datasets ; i++ )
  {
    if ( c.dataset[c.n_datasets-1].xdata )
      free( c.dataset[c.n_datasets-1].xdata );

    if ( c.dataset[c.n_datasets-1].ydata )
      free( c.dataset[c.n_datasets-1].ydata );

    if ( c.dataset[c.n_datasets-1].markers)
      free( c.dataset[c.n_datasets-1].markers);

    for( i = 0 ; i < 2 ; i++ )
    {
      if ( c.dataset[c.n_datasets-1].xerr[i])
	free( c.dataset[c.n_datasets-1].xerr[i]);

      if ( c.dataset[c.n_datasets-1].yerr[i])
	free( c.dataset[c.n_datasets-1].yerr[i]);
    }

    if ( c.dataset[c.n_datasets-1].breakval)
      free( c.dataset[c.n_datasets-1].breakval);
  }

  free( c.dataset );
}


static void
error_bar(
  double data, 		/* untransformed data (input) */
  double tdata,		/* transformed data (input) */
  double error,		/* error (input ) */
  int direction,	/* 1/-1 for data+error/data-error */
  int do_log,		/* take logarithmic transform */
  int auto_min,		/* autoscale, minimum direction */
  int auto_max,		/* autoscale, maximum direction */
  float *error_bar,	/* error bar value, relative to tdata (output) */
  float *min,		/* minimum limit (input/output) */
  float *max		/* maximum limit (input/output) */
  )
{
  double errbar;

  if ( do_log )
  {
    double err;
    if ( (err = data + direction * error ) > 0 )
    {
      errbar = log10(err);
      if ( auto_min && errbar < *min ) *min = errbar;
      if ( auto_max && errbar > *max ) *max = errbar;

      if ( tdata != -FLT_MAX )
	errbar = fabs( tdata - errbar );
    }
    else
      errbar = -FLT_MAX;

    *error_bar = errbar;
  }
  else
  {
    errbar = data + direction * error;
    if ( auto_min && errbar < *min ) *min = errbar;
    if ( auto_max && errbar > *max ) *max = errbar;
    *error_bar = error;
  }

}

/* colors to cycle through if user specifies color cycling and
   doesn't have a list */
static int color_cycle[] = { 1, 2, 3, 4, 5, 6, 8, 9 };
static int ncolor_cycle = sizeof(color_cycle) / sizeof(int);

static void
setup_dataset( DataSet* dset, int idx, int* marker_style, int* line_style, int* color )
{
  int ii;

  if ( c.color )
  {
      dset->color = c.color_cycle                   ? c.color[idx % c.n_colors]
	          : ( c.n_datasets <= c.n_colors )  ? c.color[idx]
	          :                                   c.color[c.n_colors - 1];
  }
  else
  {
      dset->color = c.color_cycle ? color_cycle[ ((*color)++ -1 ) % ncolor_cycle ]
	          :                 c.color_def;
  }

  if ( c.legend.plot )
    dset->legend_text = (c.n_datasets <= c.legend.n) ?
      c.legend.textp[idx] : c.legend.textp[c.legend.n-1];

  if ( idx < c.n_markers )
  {
    dset->marker = c.marker_attr[idx];
    *marker_style = dset->marker.style;
  }

  else
  {
    dset->marker = c.marker_attr_def;
    dset->marker.style = (*marker_style)++;
  }

  if ( INT_MAX == dset->marker.color )
  {
    int color = c.marker_attr_def.color;

    if ( idx < c.n_colors ) 
      color = c.color[idx];

    else if ( idx < c.n_connect )
      color = c.connect_attr[idx].line.color;

    if ( INT_MAX == color )
	color = dset->color;

    dset->marker.color = color;
  }

  if ( idx < c.n_connect )
  {
    dset->connect = c.connect_attr[idx];
    *line_style = c.connect_attr[idx].line.style;
  }

  else
  {
    dset->connect = c.connect_attr_def;
    dset->connect.line.style = (((*line_style)++ -1)% PGPLOT_NSTYLES) + 1;
  }

  /* the legend box code needs to know if a dataset has
     been connected; it's easiest to just set the flag here
     instead of using convoluted logic to figure it out */
  if ( Connect_Yes == dset->connect.doit )
    c.connect = 1;

  if ( INT_MAX == dset->connect.line.color )
  {
    int color = c.connect_attr_def.line.color;

    if (idx < c.n_colors ) 
      color = c.color[idx];

    if ( INT_MAX == color )
      color = dset->color;

    dset->connect.line.color = color;
  }
	  

  for ( ii = 0 ; ii < 2 ; ii++ )
  {

    if ( idx < c.n_xerr_attr )
      dset->xerr_attr[ii] = c.xerr_attr[ii][idx];

    else
      dset->xerr_attr[ii] = c.xerr_attr_def[ii];

    if ( INT_MAX == dset->xerr_attr[ii].line.color )
    {
      int color = c.xerr_attr_def[ii].line.color;

      if (idx < c.n_colors ) 
	color = c.color[idx];
	  
      else if ( idx < c.n_markers )
	color = c.marker_attr[idx].color;

      else if ( idx < c.n_connect )
	color = c.connect_attr[idx].line.color;

      if ( INT_MAX == color )
	color = dset->color;

      dset->xerr_attr[ii].line.color = color;
    }
  }

  for ( ii = 0 ; ii < 2 ; ii++ )
  {

    if ( idx < c.n_yerr_attr )
      dset->yerr_attr[ii] = c.yerr_attr[ii][idx];

    else
      dset->yerr_attr[ii] = c.yerr_attr_def[ii];

    if ( INT_MAX == dset->yerr_attr[ii].line.color )
    {
      int color = c.yerr_attr_def[ii].line.color;

      if (idx < c.n_colors ) 
	color = c.color[idx];
	  
      else if ( c.n_markers > idx )
	color = c.marker_attr[idx].color;

      else if ( c.n_connect > idx )
	color = c.connect_attr[idx].line.color;

      if ( INT_MAX == color )
	color = dset->color;

      dset->yerr_attr[ii].line.color = color;
    }
  }
}
